# -*- coding: utf-8 -*-

# module test_misc

# Copyright (C) 2019 National Research Council Canada
# Author:  Harold Parks

# This file is part of MetroloPy.

# MetroloPy is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software 
# Foundation, either version 3 of the License, or (at your option) any later 
# version.

# MetroloPy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more 
# details.

# You should have received a copy of the GNU General Public License along with 
# MetroloPy. If not, see <http://www.gnu.org/licenses/>.
"""
Check that known bugs are fixed.
"""

import metrolopy as uc
import numpy as np
import unittest


class TestMisc(unittest.TestCase):
    """
    Miscellaneaus tests.
    """
    def test_reduce(self):
        a = uc.gummy(12.367,0.22,unit='mm/m')
        a.reduce_unit()
        self.assertTrue(abs(a.x - 0.012367) < 1e-15)
        self.assertTrue(abs(a.u - 0.00022) < 1e-15)
        self.assertTrue(a.unit is uc.one)
        
    def test_zero_derivative(self):
        u = 0.00223
        a = uc.gummy(0,u)
        b = a**2
        
        self.assertTrue(b.u == 0)
        b.sim()
        self.assertTrue(abs((np.sqrt(2)*u**2-b.usim)/b.usim) < 0.1)
        b.style = 'pmsim'
        self.assertTrue('?' not in b.tostring())
        
    def test_triangular(self,plot=False):
        a = uc.gummy(uc.TriangularDist(1,1))
        a.sim()
        self.assertTrue(abs((a.usim - 1/np.sqrt(6))/a.usim) < 0.1)
        if plot:
            print('This should be a triangle with mode 1, lower limit 0 and upper limit 1:')
            a.hist()
            
        a = uc.gummy(uc.TriangularDist(-2.5,left_width=0.5,right_width=1.5))
        a.sim()
        if plot:
            print('this should be a triangle with mode -2.5, lower limit -3 and upper limit -1:')
            a.hist()
            
        a = uc.gummy(uc.TriangularDist(1,lower_limit=0.5,upper_limit=1.5))
        a.sim()
        if plot:
            print('this should be a triangle with mode 1, lower limit 0.5 and upper limit 1.5:')
            a.hist()
            
    def test_zero_degc(self):
        a = uc.gummy(0, 0.15,unit='degC')
        a.style='xf'
        self.assertTrue(a.tostring() == '0.00')
        
    def test_ufrom_multiletter(self):
        a = uc.gummy(10.0, 1, utype='A')
        b = uc.gummy(20.0, 2, utype='DUT')
        y = a - b
        self.assertTrue(abs(y.ufrom('A') - 1) < 1e-15)
        self.assertTrue(abs(y.ufrom('DUT') - 2) < 1e-15)
        
    def test_uniform_params(self):
        # issue 33
        g = uc.gummy(uc.UniformDist(lower_limit=6,upper_limit=9),unit='pA')
        g.sim()
        self.assertTrue(abs(min(g.simdata) - 6) < 0.01)
        self.assertTrue(abs(max(g.simdata) - 9) < 0.01)
        
        g = uc.gummy(uc.UniformDist(center=7.5,half_width=1.5),unit='pA')
        g.sim()
        self.assertTrue(abs(min(g.simdata) - 6) < 0.01)
        self.assertTrue(abs(max(g.simdata) - 9) < 0.01)
        
    def test_mean_b(self):
        # issue 32
        import numpy as np
        from scipy.stats import t
        from scipy import special
    
        con = special.erf(1/np.sqrt(2))
        def u(x):
            return t.ppf(1-(1-con)/2, len(x)-1)*x.std(ddof=1)/np.sqrt(len(x))
    
        x = np.random.rand(5)
        uc.gummy.bayesian = False
        g = uc.mean(x)
        g.p = 'ssd'
        self.assertTrue(abs(con-g.p) < 1e-6)
        self.assertTrue(abs(g.U - u(x)) < 1e-6)
        try:
            uc.gummy.bayesian = True
            gg = uc.mean(x)
            self.assertTrue(abs(gg.u - np.sqrt(2)*g.u) < 1e-6)
            gg.p = 'ssd'
            self.assertTrue(abs(gg.U - u(x)) < 1e-6)
        finally:
            uc.gummy.bayesian = False
            
    def test_uniform_hw(self):
        # issue 37
        a = uc.gummy(uc.UniformDist(upper_limit=1,lower_limit=-1))
        b = uc.gummy(uc.UniformDist(center=0,half_width=1))
        self.assertTrue(abs(a.u - b.u) < 1e-15)
        
        
if __name__ == '__main__':
    unittest.main()


# -*- coding: utf-8 -*-

# module test_operations

# Copyright (C) 2025 National Research Council Canada
# Author:  Harold Parks

# This file is part of MetroloPy.

# MetroloPy is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software 
# Foundation, either version 3 of the License, or (at your option) any later 
# version.

# MetroloPy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more 
# details.

# You should have received a copy of the GNU General Public License along with 
# MetroloPy. If not, see <http://www.gnu.org/licenses/>.

import metrolopy as uc
import numpy as np
import unittest

from metrolopy.tests.common import rand,make_gummy,make_number


class TestOperations(unittest.TestCase):
    """
    Tests checking basic operations with `gummy` instances using random 
    values.
    """
    def binary_func(self,f,df,sim=False,exp=None,fionly=False,uexp=-6,allowazero=True,
                    allowbzero=True,allowlargeu=True):
        if sim:
            dof = dof=float('inf')
        else:
            dof = None
        a,ax,au,adof,_,_ = make_gummy(unit=1,dof=dof,exp=exp,mpf=(not fionly),
                                      uexp=uexp,allowzero=allowazero,
                                      allowlargeu=allowlargeu)
        if rand.randint(2):
            b = make_number(exp=exp,fionly=fionly,allowzero=allowbzero)
            if isinstance(b,uc.gummy):
                bx = b.x
            else:
                bx = b
            bu = 0
            bdof = float('inf')
        else:
            b,bx,bu,bdof,_,_ = make_gummy(unit=1,dof=dof,exp=exp,mpf=(not fionly),
                                          uexp=uexp,allowzero=allowbzero,
                                          allowlargeu=allowlargeu)
            
        ax = float(ax)
        bx = float(bx)
        au = float(au)
        bu = float(bu)
        x = f(ax,bx)
        d = df(ax,bx)
        da = float(d[0])
        db = float(d[1])
        u = np.sqrt((da*au)**2 + (db*bu)**2)
        if dof is None:
            dof = ((da*au)**4/adof + (db*bu)**4/bdof)
            if dof == 0:
                dof = float('inf')
            else:
                dof = u**4/dof
                
            if dof > uc.gummy.max_dof:
                dof = float('inf')
            if dof < 1:
                dof = 1
                
            dof = float(dof)
                
        g = f(a,b)
        
        if x == 0:
            self.assertTrue(abs(g.x) < 1e-15)
        else:
            self.assertTrue(abs((g.x - x)/x) < 1e-14)
            
        if u == 0:
            self.assertTrue(abs(u) < 1e-6)
        else:
            self.assertTrue(abs((g.u - u)/u) < 1e-3)
        
        if not uc.gummy.bayesian:
            if dof is not None and not np.isinf(dof):
                self.assertTrue(abs((g.dof - dof)/dof) < 0.01)
        
        self.assertTrue(a.correlation(b) == 0)
        
        if sim and x != 0 and abs(u/x) > 1e-10:
            uc.gummy.simulate([a,b,g])
            
            self.assertTrue(abs((g.xsim - x)/(g.u/np.sqrt(1e5))) < 5)
            self.assertTrue(abs((g.u - g.usim)/g.u) < 0.1)
            
            if g.correlation(a) > 0.1:
                self.assertTrue(abs((g.correlation_sim(a)-g.correlation(a))/g.correlation(a)) < 0.1)
                
            if isinstance(b,uc.gummy) and g.correlation(b) > 0.1:
                self.assertTrue(abs((g.correlation_sim(b)-g.correlation(b))/g.correlation(b)) < 0.1)
        
    def test_add(self,n=1000,sim=False):
        """
        Test gummy addition.
        """
        def f(a,b):
            return a + b
        def df(ax,bx):
            return (1,1)
        for i in range(n):
            self.binary_func(f,df,sim=sim)
            
    def test_sub(self,n=1000,sim=False):
        """
        Test gummy subtraction.
        """
        def f(a,b):
            return a - b
        def df(ax,bx):
            return (1,-1)
        for i in range(n):
            self.binary_func(f,df,sim=sim)
            
    def test_mul(self,n=1000,sim=False):
        """
        Test gummy multiplication.
        """
        def f(a,b):
            return a*b
        def df(ax,bx):
            return (bx,ax)
        for i in range(n):
            self.binary_func(f,df,sim=sim)
            
    def test_div(self,n=1000,sim=False):
        """
        Test gummy division.
        """
        def f(a,b):
            return a/b
        def df(ax,bx):
            return (1/bx,-ax/bx**2)
        if sim:
            lu = False
        else:
            lu = True
        for i in range(n):
            self.binary_func(f,df,sim=sim,allowbzero=False,allowlargeu=lu)
            
    def test_pow(self,n=1000,sim=False):
        """
        Test gummy exponentiation.
        """
        def f(a,b):
            return abs(a)**b
        def df(ax,bx):
            ax = abs(ax)
            return (bx*ax**(bx-1),np.log(ax)*ax**bx)
        if sim:
            lu = False
        else:
            lu = True
        for i in range(n):
            self.binary_func(f,df,sim=sim,exp=0,allowazero=False,allowlargeu=lu)
            
    def test_mod(self,n=1000,sim=False):
        """
        Test gummy mod.
        """
        def f(a,b):
            return a%b
        def df(ax,bx):
            return (1,np.sign(bx)*abs(ax//bx))
        for i in range(n):
            self.binary_func(f,df,sim=sim,fionly=True,allowazero=False)
        
    def test_abs(self,n=1000,sim=False):
        """
        Test gummy abs.
        """
        def f(a,b):
            return abs(a)*abs(b)
        def df(ax,bx):
            return (abs(bx),abs(ax))
        if sim:
            lu = False
        else:
            lu = True
        for i in range(n):
            self.binary_func(f,df,sim=sim,allowlargeu=lu,allowazero=lu,
                        allowbzero=lu)
            
    def test_neg(self,n=1000,sim=False):
        """
        Test gummy neg.
        """
        def f(a,b):
            return (-a)*b
        def df(ax,bx):
            return (-bx,-ax)
        for i in range(n):
            self.binary_func(f,df,sim=sim)
            
    def test_pos(self,n=1000,sim=False):
        """
        Test gummy pos.
        """
        def f(a,b):
            return (+a)*b
        def df(ax,bx):
            return (bx,ax)
        for i in range(n):
            self.binary_func(f,df,sim=sim)
            
    def test_sincos(self,n=1000,sim=False):
        """
        Test gummy metrolopy.sin + metrolopy.cos.
        """
        def f(a,b):
            return uc.sin(a) + uc.cos(b)
        def df(ax,bx):
            return (uc.cos(ax),-uc.sin(bx))
        for i in range(n):
            self.binary_func(f,df,sim=sim)
            
    def test_npsincos(self,n=1000,sim=False):
        """
        Test gummy numpy.sin + numpy.cos.
        """
        def f(a,b):
            return np.sin(a) + np.cos(b)
        def df(ax,bx):
            return (np.cos(ax),-np.sin(bx))
        for i in range(n):
            self.binary_func(f,df,sim=sim,fionly=True)
            
    def test_ap1sincos(self,n=1000,sim=False):
        """
        Test gummy.apply(numpy.sin) + gummy.apply(numpy.cos).
        """
        def f(a,b):
            return uc.gummy.apply(np.sin,np.cos,a) + uc.gummy.napply(np.cos,b)
        def df(ax,bx):
            return (np.cos(ax),-np.sin(bx))
        for i in range(n):
            self.binary_func(f,df,sim=sim,fionly=True,exp=0,allowlargeu=False)
            
    def test_ap2sincos(self,n=1000,sim=False):
        """
        Test gummy.apply(numpy.sin + numpy.cos).
        """
        def ff(a,b):
            return np.sin(a) + np.cos(b)
        def df(ax,bx):
            return (np.cos(ax),-np.sin(bx))
        def f(a,b):
            return uc.gummy.apply(ff,df,a,b)
    
        for i in range(n):
            self.binary_func(f,df,sim=sim,fionly=True)
            
    def test_apnsincos(self,n=1000,sim=False):
        """
        Test gummy.napply(numpy.sin + numpy.cos).
        """
        def ff(a,b):
            return np.sin(a) + np.cos(b)
        def df(ax,bx):
            return (np.cos(ax),-np.sin(bx))
        def f(a,b):
            return uc.gummy.napply(ff,a,b)
    
        for i in range(n):
            self.binary_func(f,df,sim=sim,fionly=True,exp=0,allowlargeu=False)
            
    def test_addxmul(self,n=1000):
        """
        Check that x+x+x+...+x == n*x for gummy values.
        """
        def f(*x):
            r = x[0]
            for i in range(len(x)-1):
                r += x[i+1]
            return r
            
        for j in range(n):
            k = rand.randint(10) + 2
            x = make_gummy(unit=1)[0]
            a = f(*(k*[x]))
            b = k*x
            c = uc.gummy.apply(f,lambda *x: len(x)*[1],*(k*[x]))
            d = uc.gummy.napply(f,*(k*[x]))
            
            if b.x == 0:
                self.assertTrue(abs(a.x - b.x) < 1e-14)
            else:
                self.assertTrue(abs((a.x - b.x)/b.x) < 1e-14)
            if x.u == 0:
                self.assertTrue(a.u == 0)
                self.assertTrue(b.u == 0)
            else:
                self.assertTrue((a.u - b.u)/b.u < 1e-3)
                self.assertTrue(a.correlation(b) == 1)
                self.assertTrue(a.correlation(x) == 1)
            
            if b.x == 0:
                self.assertTrue(abs(c.x - b.x) < 1e-14)
            else:
                self.assertTrue(abs((c.x - b.x)/b.x) < 1e-14)
            if x.u == 0:
                self.assertTrue(c.u == 0)
            else:
                self.assertTrue((c.u - b.u)/b.u < 1e-3)
                self.assertTrue(c.correlation(b) == 1)
            
            if b.x == 0:
                self.assertTrue(abs(d.x - b.x) < 1e-14)
            else:
                self.assertTrue(abs((d.x - b.x)/b.x) < 1e-14)
            if x.u == 0:
                self.assertTrue(d.u == 0)
            else:
                self.assertTrue(abs((d.u - b.u)/b.u) < 1e-3)
                self.assertTrue(abs(d.correlation(b) - 1) < 1e-14)
            
    def test_mulxpow(self,n=1000):
        """
        Check that x*x*x*...*x == x**n for gummy values.
        """
        def f(*x):
            r = x[0]
            for i in range(len(x)-1):
                r *= x[i+1]
            return r
            
        for j in range(n):
            k = rand.randint(10) + 2
            x = make_gummy(unit=1)[0]
            a = f(*(k*[x]))
            b = x**k
            xx = x.x
            c = uc.gummy.apply(f,lambda *x: len(x)*[xx**(k-1)],*(k*[x]))
            d = uc.gummy.napply(f,*(k*[x]))
            
            if b.x == 0:
                self.assertTrue(abs(a.x - b.x) < 1e-14)
            else:
                self.assertTrue(abs((a.x - b.x)/b.x) < 1e-14)
            if x.u == 0:
                self.assertTrue(a.u == 0)
                self.assertTrue(b.u == 0)
            else:
                self.assertTrue(abs((a.u - b.u)/b.u) < 1e-3)
                self.assertTrue(a.correlation(b) == 1)
            
            if k%2 == 0 and x.x < 0:
                self.assertTrue(a.correlation(x) == -1)
            else:
                self.assertTrue(a.correlation(x) == 1)
                
            if b.x == 0:
                self.assertTrue(abs(c.x - b.x) < 1e-14)
            else:
                self.assertTrue(abs((c.x - b.x)/b.x) < 1e-14)
            if x.u == 0:
                self.assertTrue(c.u == 0)
            else:
                self.assertTrue(abs((c.u - b.u)/b.u) < 1e-3)
                self.assertTrue(c.correlation(b) == 1)
            
            if b.x == 0:
                self.assertTrue(abs(d.x - b.x) < 1e-14)
            else:
                self.assertTrue(abs((d.x - b.x)/b.x) < 1e-14)
            if x.u == 0:
                self.assertTrue(d.u == 0)
            else:
                self.assertTrue(abs((d.u - b.u)/b.u) < 1e-3)
                self.assertTrue(d.correlation(b) == 1)
    
            
    def test_mulxnpow(self,n=1000):
        """
        Check that (1/x)*(1/x)*(1/x)*...*(1/x) == x**-n for gummy values.
        """
        def f(*x):
            r = 1/x[0]
            for i in range(len(x)-1):
                r *= 1/x[i+1]
            return r
            
        for j in range(n):
            k = rand.randint(10) + 2
            x = make_gummy(unit=1,allowzero=False,allowlargeu=False)[0]
            a = f(*(k*[x]))
            b = x**-k
            d = uc.gummy.napply(f,*(k*[x]))
            
            if b.x == 0:
                self.assertTrue(abs(a.x - b.x) < 1e-14)
            else:
                self.assertTrue(abs((a.x - b.x)/b.x) < 1e-14)
            if x.u == 0:
                self.assertTrue(a.u == 0)
                self.assertTrue(b.u == 0)
            else:
                self.assertTrue(abs((a.u - b.u)/b.u) < 1e-3)
                self.assertTrue(a.correlation(b) == 1)
            
            if k%2 == 0 and x.x < 0:
                self.assertTrue(a.correlation(x) == 1)
            else:
                self.assertTrue(a.correlation(x) == -1)
            
            if b.x == 0:
                self.assertTrue(abs(d.x - b.x) < 1e-14)
            else:
                self.assertTrue(abs((d.x - b.x)/b.x) < 1e-14)
            if x.u == 0:
                self.assertTrue(d.u == 0)
            else:
                self.assertTrue(abs((d.u - b.u)/b.u) < 1e-3)
                self.assertTrue(d.correlation(b) == 1)
                
                
if __name__ == '__main__':
    unittest.main()


# -*- coding: utf-8 -*-

# module test_complex

# Copyright (C) 2019 National Research Council Canada
# Author:  Harold Parks

# This file is part of MetroloPy.

# MetroloPy is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software 
# Foundation, either version 3 of the License, or (at your option) any later 
# version.

# MetroloPy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more 
# details.

# You should have received a copy of the GNU General Public License along with 
# MetroloPy. If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import unittest
from metrolopy import immy,ummy,jummy,gummy
from metrolopy.tests.common import display

rand = np.random.RandomState()

def make_immy(prnt=False):
    if rand.randint(2):
        imake = immy
        make = ummy
    else:
        imake = jummy
        make = gummy
    r = 4*rand.rand() + 1
    if rand.randint(2):
        r = -r
    ru = (0.1*rand.rand() + 0.01)*(abs(r))
    i = 4*rand.rand() + 1
    if rand.randint(2):
        i = -i
    iu = (0.1*rand.rand() + 0.01)*(abs(i))
    c = (2*rand.rand() - 1)*ru*iu

    cov = [[ru**2,c],[c,iu**2]]
    real,imag = make.create([r,i],covariance_matrix=cov)
    
    if rand.randint(2):
        ret = imake(real=real,imag=imag)
    else:
        if rand.randint(2):
            ret = imake(real=r,imag=i,cov=cov)
            assert abs(ret.real.covariance(ret.imag) - cov[0][1]) < 1e-6
            if imake is immy:
                real._ref = ret.real._ref
                imag._ref = ret.imag._ref
            else:
                real.value._ref = ret.real.value._ref
                imag.value._ref = ret.imag.value._ref
        else:
            r = np.sqrt(real**2 + imag**2)
            phi = np.arctan2(imag,real)
            ret = imake(r=r.x,phi=phi.x,cov=make.covariance_matrix([r,phi]))
            assert abs(ret.real.covariance(ret.imag) - cov[0][1]) < 1e-6
            if imake is immy:
                real._ref = ret.real._ref
                imag._ref = ret.imag._ref
            else:
                real.value._ref = ret.real.value._ref
                imag.value._ref = ret.imag.value._ref
                
    if prnt:
        display(ret)
    
    return (ret,real,imag)


class TestComplex(unittest.TestCase):
    """
    Tests for the `immy` and `jummy` classes.  Basic operations and
    instance creation are tested using random values.
    """
    def assert_ummy_close(self,u1,u2):
            self.assertTrue(abs(u1.correlation(u2)) > 1 - 1e-4)
            u1x = max(u1.x,u1.u,u2.x,u2.u)
            self.assertTrue(abs((u1.x - u2.x)/(u1x)) < 1e-10)
            self.assertTrue(abs((u1.u - u2.u)/(u1.u)) < 1e-2)
            
            if u1.dof == float('inf'):
                self.assertTrue(u2.dof == float('inf'))
            else:
                self.assertTrue(abs((u2.dof - u1.dof)/u1.dof) < 1e-2)
            
    def assert_immy_close(self,i1,i2):
            self.assert_ummy_close(i1.real,i2.real)
            self.assert_ummy_close(i1.imag,i2.imag)
    
    def test_immy_init(self,n=1000,prnt=False):
        """
        Test immy creation using the real,imag,r and phi keywords.
        """
        for m in range(n):
            x,xr,xi = make_immy(prnt=prnt)
            
            if rand.randint(2):
                if rand.randint(2):
                    x = type(x)(x)
                else:
                    x = x.copy(formatting=False)
            
            self.assert_ummy_close(x.real,xr)
            self.assert_ummy_close(x.imag,xi)
            self.assert_ummy_close(x.angle(),np.arctan2(xi,xr))
            self.assert_ummy_close(abs(x),np.sqrt(xr**2 + xi**2))
            self.assert_immy_close(x.conjugate(),immy(real=xr,imag=-xi))
            self.assert_immy_close(-x,immy(real=-xr,imag=-xi))
            self.assert_immy_close(+x,immy(real=xr,imag=xi))
            
            if prnt:
                if rand.randint(2):
                    y =1e12*x
                else:
                    y = x/1e12
                display(y)
    
    def _test_immy_bop(self,f,nf,n=1000,prnt=False,allow_small=True):
        m = 0
        while m < n:
            a,ar,ai = make_immy()
            if True:#rand.randint(2):
                b,br,bi = make_immy()
            else:
                if rand.randint(2):
                    b = 4*rand.rand() + 1
                    if rand.randint(2):
                        b = -b
                    if rand.randint(2):
                        bu = (0.1*rand.rand()+0.01)*(abs(b))
                        b = ummy(b,u=bu)
                else:
                    br = 4*rand.rand() + 1
                    if rand.randint(2):
                        br = -br
                    bi = 4*rand.rand() + 1
                    if rand.randint(2):
                        bi = -bi
                    b = complex(br,bi)
                if rand.randint(2):
                    bb = b
                    b = a
                    a = bb
                    
            if isinstance(a,(immy,ummy)):
                ax = a.x
            else:
                ax = a
            if isinstance(b,(immy,ummy)):
                bx = b.x
            else:
                bx = b
            
                
            cx = f(ax,bx)
            
            if allow_small or abs(cx) > 0.1:
                m +=1
                
                c = f(a,b)
                cn = type(c).napply(nf,a,b)
                
                if prnt:
                    display(a)
                    display(b)
                    display(c)
                    display(cn)
                    print('---')
        
                self.assertTrue(abs((c.real.x - cx.real)/cx.real) < 1e-10)
                self.assertTrue(abs((c.imag.x - cx.imag)/cx.imag) < 1e-10)
                self.assert_immy_close(c,cn)
        
    def test_immy_add(self,n=1000,prnt=False):
        """
        Test immy addition.
        """
        self._test_immy_bop(lambda a,b: a + b,np.add,n,prnt)
        
        for m in range(10):
            i = make_immy()[0]
            self.assertTrue(i + 0 == i)
            self.assertTrue(0 + i == i)
        
    def test_immy_sub(self,n=1000,prnt=False):
        """
        Test immy subtraction.
        """
        self._test_immy_bop(lambda a,b: a - b,np.subtract,n,prnt)
        
        for m in range(10):
            i = make_immy()[0]
            self.assertTrue(i - 0 == i)
            self.assertTrue(0 - i == -i)
        
    def test_immy_mul(self,n=1000,prnt=False):
        """
        Test immy multiplication.
        """
        self._test_immy_bop(lambda a,b: a*b,np.multiply,n,prnt)
        
        for m in range(10):
            i = make_immy()[0]
            self.assertTrue(i*0 == 0)
            self.assertTrue(i*immy(0) == 0)
            self.assertTrue(i*1 == i)
            self.assertTrue(i*immy(1) == i)
            self.assertTrue(1*i == i)
            self.assertTrue(immy(1)*i == i)
        
    def test_immy_div(self,n=1000,prnt=False):
        """
        Test immy division.
        """
        self._test_immy_bop(lambda a,b: a/b,np.divide,n,prnt,allow_small=False)
        
        for m in range(10):
            i = make_immy()[0]
            self.assertTrue(i/1 == i)
        
    def test_immy_pow(self,n=1000,prnt=False):
        """
        Test immy exponentiation.
        """
        #_test_immy_bop(lambda a,b: a**b,np.power,n,prnt,allow_small=False)
        
        for m in range(10):
            i = make_immy()[0]
            self.assertTrue(0**i == immy(0))
            self.assertTrue(0**i == ummy(0))
            self.assertTrue(0**i == 0)
            self.assertTrue(i**0 == 1)
            
            self.assert_immy_close(i**-1,1/i)
            self.assert_immy_close(i**2,i*i)
            self.assert_immy_close(i**3,i*i*i)
    
if __name__ == '__main__':
    unittest.main()


# -*- coding: utf-8 -*-

# module miscunits

# Copyright (C) 2025 National Research Council Canada
# Author:  Harold Parks

# This file is part of MetroloPy.

# MetroloPy is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software 
# Foundation, either version 3 of the License, or (at your option) any later 
# version.

# MetroloPy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more 
# details.

# You should have received a copy of the GNU General Public License along with 
# MetroloPy. If not, see <http://www.gnu.org/licenses/>.

"""
This module is loaded by the gummy.units module and is not intended be be
imported directly.  

Miscellaneous unit definitions: astronomical units, natural units, some 
non-SI units accepted for use with the SI not in the siunits module, some 
obsolete SI units, and a few other units.
"""

from numpy import sqrt,log,log2,log10
from .gummy import gummy
from .unit import Unit,Conversion,MFraction
from .prefixedunit import PrefixedUnit,BinaryPrefixedUnit
from .logunit import LogUnit,LogConversion
from .siunits import (_kg,_V,_ohm,_s,_K,_J,_m,_d,_min,_h,_W,_Hz,_C,_A,_N,_cm,
                      _Wb,_T,_Pa)
from .constcom import (dalton,KJ,RK,KJ90,RK90,e,c,solar_mass,jupiter_mass,
                       earth_mass,hbar,a0,me,mp,G,alph,k,pi,euler)

def _lg10(x):
    try:
        from mpmath import mpf
        from mpmath import log10 as mlog10
    except:
        mlog10 = mpf = None
        
    if isinstance(x,mpf):
        return mlog10(x)
    if hasattr(x,'log10'):
        return x.log10()
    try:
        return log10(x)
    except:
        return log10(float(x))

with Unit._builtin():
    
    # non-SI accepted for use with the SI, SI Brochure, 9th ed
    _Da = PrefixedUnit('dalton','Da',Conversion(_kg,dalton),
                       add_symbol=True,description='unified atomic mass unit')
    
    PrefixedUnit('electronvolt','eV',Conversion(_J,e),add_symbol=True,order=0,
                 description='unit of energy, the energy gained by one electron moving across one volt')


    #Other non-SI units
    PrefixedUnit('bar','bar',Conversion(_Pa,100000),add_symbol=True,order=0,
                 description='unit of pressure')
    Unit('millimetre of mercury','mmHg',Conversion(_Pa,133.322387415),add_symbol=True,
         order=0,description='unit of pressure')
    Unit('angstrom','\u212B',Conversion(_m,MFraction('1e-10')),add_symbol=True,order=0,
         html_symbol='&#8491;',ascii_symbol='angstrom',description='unit of length')
    
    _nm = Unit('nautical mile','M',Conversion(_m,1852),add_symbol=True,order=0,
               description='unit of length')
    Unit.alias('NM',_nm)
    Unit.alias('nmi',_nm)
    Unit.alias('Nm',_nm)
    
    Unit('barn','b',Conversion(_m**2,MFraction('1e-28')),add_symbol=True,order=0,
         description='unit of area')
    Unit('knot','kn',Conversion(_nm*_h**-1,1),add_symbol=True,order=0,
         description='unit of speed')
    _dbp = LogUnit('decibel','dB',LogConversion(1,10,10,_lg10),add_symbol=True)
    Unit.alias('dB power',_dbp)
    Unit.alias('dB-p',_dbp)
    
    _dbf = LogUnit('decibel field','dB',LogConversion(1,20,10,_lg10),short_name='dB-f')
    Unit.alias('dB field',_dbf)
    Unit.alias('dB root-power',_dbf)
    
    def _ln(x):
        try:
            return log(x)
        except:
            return log(float(x))
    LogUnit('neper','Np',LogConversion(1,1,euler,_ln),add_symbol=True)


    # astronomical units
    _c = Unit('speed of light','c',Conversion(_m/_s,c),add_symbol=True,order=0,
         html_symbol='<i>c</i>',latex_symbol='\t\tc',
         description='natural unit of velocity')
    _au = Unit('astronomical unit','au',Conversion(_m,149597870700),add_symbol=True,
               order=0,description='astronomical unit of length')
    Unit.alias('ua',_au)
    PrefixedUnit('parsec','pc',Conversion(_au,648000/pi),add_symbol=True,order=0,
                 prefixes=['kilo','mega','giga'],
                 description='astronomical unit of length')
    _a = PrefixedUnit('Julian year','a',Conversion(_d,MFraction('365.25')),add_symbol=True,order=0,
                       description='astronomical unit of time',prefixes=['kilo','mega',
                       'giga','tera'],additional_names=('annum','year'),
                        additional_short_names=('yr',))
    Unit('light year','ly',Conversion(_c*_a,1),add_symbol=True,order=0,
         description='astronomical unit of length')
    Unit('siriometer','Sm',Conversion(_au,1000000),add_unit=False,
         description='astronomical unit of length')
    Unit('light second','light-second',Conversion(_c*_s,1),add_symbol=True,
         order=0,description='unit of length',)
    Unit('light minute','light-minute',Conversion(_c*_min,1),add_symbol=True,
         order=0,description='astronomical unit of length')
    Unit('light hour','light-hour',Conversion(_c*_h,1),add_symbol=True,
         order=0,description='astronomical unit of length')
    _M_solar = Unit('solar mass','M\u2609',Conversion(_kg,solar_mass),
                    add_symbol=True,order=0,html_symbol='<i>M</i><sub>&#x2609;</sub>',
                    latex_symbol='\t\tM_{\\odot}',ascii_symbol='M(solar)',
                    description='astronomical unit of mass, mass of sun')
    Unit.alias('M(Sun)',_M_solar)
    _M_J = Unit('Jupiter mass','M(J)',Conversion(_kg,jupiter_mass),add_symbol=True,
         order=0,html_symbol='<i>M</i><sub>J</sub>',latex_symbol='\t\tM_{\\mathrm{J}}',
         description='astronomical unit of mass') #check
    Unit.alias('M(Jup)',_M_J)
    Unit('Earth mass','M\u2295',Conversion(_kg,earth_mass),add_symbol=True,
         order=0,html_symbol='<i>M</i><sub>&#x2295;</sub>',latex_symbol='\t\tM_{\\oplus}',
         ascii_symbol='M(E)',description='astronomical unit of mass')
    _Jy = Unit('jansky','Jy',Conversion(_W*_m**-2*_Hz**-1,MFraction('1e-26')),add_symbol=True,
         decription='astronomical unit, spectral flux density, spectral irradiance')
    LogUnit('monochromatic AB magnitude','m(AB)',
            LogConversion(gummy(3631,unit=_Jy),-2.5,10,_lg10),
            add_symbol=True,html_symbol='<i>m</i><sub>AB</sub>',
            latex_symbol='\t\tm_{\\mathrm{AB}}',
            description='astronomical unit, spectral flux density, spectral irradiance')
    
    # natural units
    _hbar = Unit('natural unit of action','\u210f',Conversion(_J*_s,hbar),
         order=0,add_symbol=True,ascii_symbol='h-bar')
    _e = Unit('elementary charge','e',Conversion(_C,e),
         order=0,add_symbol=True,html_symbol='<i>e</i>',latex_symbol='\t\te',
         description='natural unit of charge, natural unit, atomic unit of charge')
    _a0 = Unit('bohr','a0',Conversion(_m,a0),
         order=0,add_symbol=True,html_symbol='<i>a</i><sub>0</sub>',latex_symbol='\t\ta_{0}',
         description='natural unit of length, atomic unit of length, Bohr radius')
    Unit.alias('a(0)',_a0)
    
    _me = Unit('electon mass','m(e)',Conversion(_kg,me),
         order=0,add_symbol=True,html_symbol='<i>m</i><sub>e</sub>',latex_symbol='\t\tm_{\\mathrm{e}}',
         description='natural, atomic unit of mass')
    _Eh = Unit('hartree','E_h',Conversion(_hbar**2*_me**-1*_a0**-2,1),order=0,
               add_symbol=True,html_symbol='<i>E</i><sub>h</sub>',latex_symbol='\t\tE_{\\mathrm{h}}',
               description='natural, atomic unit of energy')
    Unit.alias('Ha',_Eh)
    Unit('reduced Compton wavelength','\u019b(C)',
         Conversion(_hbar*_me**-1*_c**-1,1),order=0,add_symbol=True,
         html_symbol='\u019b<sub>C</sub>',latex_symbol='\t\t\u019b_{\\mathrm{C}}',
         ascii_symbol='lambda(C)',
         description='electron reduced compton wavelength, natural unit of length')
    Unit('proton mass','m(p)',Conversion(_kg,mp),
         order=0,add_symbol=True,html_symbol='<i>m</i><sub>p</sub>',latex_symbol='\t\tm_{\\mathrm{p}}')

    Unit('Planck length','l(P)',Conversion(_m,sqrt(hbar*G*c**-3)),order=0,
         add_symbol=True,html_symbol='<i>l</i><sub>P</sub>',
         latex_symbol='\t\tl_{\\mathrm{P}}',description='natural unit of length')
    Unit('Planck mass','m(P)',Conversion(_kg,sqrt(hbar*G**-1*_c)),order=0,
         add_symbol=True,html_symbol='<i>m</i><sub>P</sub>',
         latex_symbol='\t\tm_{\\mathrm{P}}',description='natural unit of mass')
    Unit('Planck time','t(P)',Conversion(_s,sqrt(hbar*G*c**-5)),order=0,
         add_symbol=True,html_symbol='<i>t</i><sub>P</sub>',
         latex_symbol='\t\tt_{\\mathrm{P}}',description='natural unit of time')
    Unit('Planck charge','q(P)',Conversion(_C,e/sqrt(alph)),order=0,
         add_symbol=True,html_symbol='<i>q</i><sub>P</sub>',
         latex_symbol='\t\tq_{\\mathrm{P}}',description='natural unit of charge')
    Unit('Planck temperature','T(P)',Conversion(_K,sqrt(hbar*G**-1*c**5*k**-2)),order=0,
         add_symbol=True,html_symbol='<i>T</i><sub>P</sub>',
         latex_symbol='\t\tT_{\\mathrm{P}}',description='natural unit of temperature')
    
    PrefixedUnit('volt ampere reactive','var',Conversion(_V*_A,1),add_symbol=True,
                 order=0,description='volt-ampere reactive power')
    PrefixedUnit('volt ampere','VA',Conversion(_V*_A,1),add_symbol=True,
                 order=0,description='volt-ampere apparent power')
    
    
    #Non-SI units associated with the CGS and the CGS-Gaussian system of units
    PrefixedUnit('erg','erg',Conversion(_J,MFraction('1e-7')),add_symbol=True,order=0,
                 description='CGS unit of energy')
    _dyn = PrefixedUnit('dyne','dyn',Conversion(_N,MFraction('1e-5')),add_symbol=True,order=0,
                        description='CGS unit of force')
    PrefixedUnit('poise','P',Conversion('dyn s cm**-2',1),add_symbol=True,order=0,
                 description='CGS unit of dynamic viscosity')
    PrefixedUnit('stokes','St',Conversion('cm**2 s**-1',1),add_symbol=True,order=0,
                 description='CGS unit of kinematic viscosity')
    PrefixedUnit('stilb','sb',Conversion('cd cm**-2',1),add_symbol=True,order=0,
                 description='CGS unit of luminance')
    PrefixedUnit('phot','ph',Conversion('cd sr cm**-2',1),add_symbol=True,order=0,
                 description='CGS unit of illuminance or luminous flux')
    PrefixedUnit('galileo','Gal',Conversion('cm s**-2',1),add_symbol=True,order=0,
                 description='CGS unit of acceleration')
    PrefixedUnit('maxwell','Mx',Conversion(_Wb,MFraction('1e-8')),add_symbol=True,order=0,
                 description='CGS unit of magnetic flux')
    PrefixedUnit('gauss','G',Conversion(_T,MFraction('1e-4')),add_symbol=True,order=0,
                 description='CGS unit of magnetic flux density')
    PrefixedUnit('oersted','Oe',Conversion(_A*_m**-1,1000/(4*pi)),add_symbol=True,
                 order=0,description='CGS unit of auxiliary magnetic field')
    
    
    # decibel units
    _dBSPL = LogUnit('decibel sound pressure level','dB',
                 LogConversion(gummy(20,unit='uPa'),20,10,_lg10),
                 short_name='dB(SPL)',add_symbol=False,
                 description='sound pressure level in air')
    Unit.alias('dB(20uPa)',_dBSPL)
    
    _dBuPa = LogUnit('decibel \u03BCPa','dB',
                   LogConversion(gummy(1,unit='uPa'),20,10,_lg10),
                   short_name='dB(uPa)',add_symbol=False,
                   description='sound pressure level in water')
    
    _dBSIL = LogUnit('decibel sound intensity level','dB',
                 LogConversion(gummy(1e-12,unit=_W/_m**2),10,10,_lg10),
                 short_name='dB(SIL)',add_symbol=False)
    
    _dBSWL = LogUnit('decibel sound power level','dB',
                 LogConversion(gummy(1e-12,unit=_W),10,10,_lg10),
                 short_name='dB(SWL)',add_symbol=False)
    Unit.alias('dB(pW)',_dBSWL)
    
    _dBm = LogUnit('decibel milliwatt','dBm',LogConversion(gummy(1,unit='mW'),10,10,_lg10),
                   add_symbol=True,description='unit of power')
    Unit.alias('dB(mW)',_dBm)
    
    _dBV = LogUnit('decibel volt','dBV',LogConversion(gummy(1,unit=_V),20,10,_lg10),
                   add_symbol=True,description='unit of voltage')
    Unit.alias('dB(V)',_dBV)
    
    _dBv = LogUnit('decibel u','dBu',LogConversion(gummy(sqrt(0.6),unit=_V),20,10,_lg10),
                   add_symbol=True,
                   description='unit of voltage, volt, decibel relative to sqrt(0.6) V')
    Unit.alias('dBu',_dBv)
    
    _dBmV = LogUnit('decibel millivolt','dBmV',LogConversion(gummy(1,unit='mV'),20,10,_lg10),
                    add_symbol=True,description='unit of voltage, volt')
    Unit.alias('dB(mV)',_dBmV)
    
    _dBuV = LogUnit('decibel microvolt','dB\u03bcV',LogConversion(gummy(1,unit='uV'),20,10,_lg10),
                    add_symbol=True,ascii_symbol='dBuV',description='unit of voltage')
    Unit.alias('dB(uV)',_dBuV)
    
    _dBZ = LogUnit('decibel Z','dBZ',LogConversion(gummy(1,unit='mm**6 m**-3'),10,10,_lg10),
                    add_symbol=True,description='unit used with weather radar')
    Unit.alias('dB(Z)',_dBZ)
    
    _dBJ = LogUnit('decibel joule','dBJ',LogConversion(gummy(1,unit=_J),10,10,_lg10),
                    add_symbol=True,description='unit of energy')
    Unit.alias('dB(J)',_dBJ)
    
    _dBmu = LogUnit('decibel microvolt per metre','dB\u03bc',LogConversion(gummy(1,unit='uV/m'),20,10,_lg10),
                    add_symbol=True,description='unit of voltage')
    Unit.alias('dB(uV/m)',_dBmu)
    Unit.alias('decibel microvolt per meter',_dBmu)
    
    _dBf = LogUnit('decibel femtowatt','dBf',LogConversion(gummy(1,unit='fW'),10,10,_lg10),
                    add_symbol=True,description='unit of power')
    Unit.alias('dB(fW)',_dBf)
    
    _dBW = LogUnit('decibel watt','dBW',LogConversion(gummy(1,unit=_W),10,10,_lg10),
                    add_symbol=True,description='unit of power')
    Unit.alias('dB(W)',_dBW)
    
    _dBk = LogUnit('decibel kilowatt','dBk',LogConversion(gummy(1,unit='kW'),10,10,_lg10),
                    add_symbol=True,description='unit of power')
    Unit.alias('dB(kW)',_dBk)
    
    _dBsm = LogUnit('decibel square metre','dBsm',LogConversion(gummy(1,unit=_m**2),10,10,_lg10),
                    add_symbol=True,description='unit used to measure antenna effective area')
    Unit.alias('dB(m**2)',_dBsm)
    Unit.alias('decibel square meter',_dBsm)
    
    _dBmm1 = LogUnit('decibel reciprocal metre','dB(m\u207b\u00b9)',LogConversion(gummy(1,unit=_m**-1),10,10,_lg10),
                    add_symbol=True,html_symbol='dB(m<sup>-1</sup>)',
                    latex_symbol='\t\t\\mathrm{dB}(\\mathrm{m}^{-1})',
                    ascii_symbol='dB(m**-1)',
                    description='unit used to measure antenna factor')
    Unit.alias('decibel reciprocal meter',_dBmm1)
    
    _dBHz = LogUnit('decibel hertz','dB-Hz',LogConversion(gummy(1,unit=_Hz),10,10,_lg10),
                    add_symbol=True,description='frequency')
    Unit.alias('dB(Hz)',_dBHz)
    
    _dBK = LogUnit('decibel kelvin','dBK',LogConversion(gummy(1,unit=_K),10,10,_lg10),
                    add_symbol=True,description='unit used to measure noise temperature')
    Unit.alias('dB(K)',_dBK)
    
    _dBmm1 = LogUnit('decibel reciprocal kelvin','dB(K\u207b\u00b9)',LogConversion(gummy(1,unit=_K**-1),20,10,_lg10),
                    add_symbol=True,html_symbol='dB(K<sup>-1</sup>)',
                    latex_symbol='\t\t\\mathrm{dB}(\\mathrm{K}^{-1})',
                    ascii_symbol='dB(K**-1)',description='temperature')
    
    _mBm = LogUnit('millibel milliwatt','mBm',LogConversion(gummy(1,unit='mW'),1000,10,_lg10),
                    add_symbol=True,description='unit of power')
    Unit.alias('mB(mW)',_mBm)
    
    LogUnit('bel','B',LogConversion(1,1,10,_lg10),add_symbol=False)
    
    _bit = BinaryPrefixedUnit('bit','bit',order=0,add_symbol=True,
                              additional_names=('shannon',),
                              description='unit of information')
    Unit.alias('shannon',_bit)
    
    _nat = Unit('natural unit of information','nat',Conversion(_bit,1/log(2)),
         add_symbol=True,order=0,description='unit of information')
    Unit.alias('nit',_nat)
    Unit.alias('nepit',_nat)
    
    _Hart = Unit('hartley','Hart',Conversion(_bit,log2(10)),add_symbol=True,
                 order=0,description='unit of information')
    Unit.alias('ban',_Hart)
    Unit.alias('dit',_Hart)
    
    BinaryPrefixedUnit('byte','B',Conversion(_bit,8),add_symbol=True,order=0,
                       description='unit of information')
    
    _nibble = Unit('nibble','nibble',Conversion(_bit,4),add_symbol=True,order=0,
                   description='unit of information')
    Unit.alias('nybble',_nibble)
    Unit.alias('nyble',_nibble)
    
    PrefixedUnit('torr','Torr',Conversion(_Pa,MFraction(101325,760)),add_symbol=True,
                         order=0,prefixes=['milli'],description='unit of pressure')
    Unit('standard atmosphere','atm',Conversion(_Pa,101325),add_symbol=True,order=0,description='unit of pressure')
                 
    _M_W = LogUnit('moment magnitude','M(W)',
                   LogConversion(gummy(1,unit=_dyn*_cm),MFraction(2,3),10,_lg10,offset=-10.7),
                   add_symbol=True,html_symbol='<i>M</i><sub>W</sub>',
                   latex_symbol='\\textit{M}_{W}',
                   desription='moment magnitude scale, earthquake intensity')
    Unit.alias('MMS',_M_W)
    
    Unit.alias('wavenumber',_cm**-1)
    
    Unit('root hertz','\u221aHz',Conversion(_Hz**0.5,1),add_symbol=True,order=0,
         html_symbol='&radic;<span style="text-decoration:overline;">Hz</span>',
         latex_symbol='\t\t\\sqrt{\\mathrm{Hz}}',ascii_symbol='sqrtHz')


    # 1990 conventional unit definitions
    _V90 = PrefixedUnit('volt 90','V(90)',Conversion(_V,KJ90/KJ),add_symbol=True,
                        html_symbol='V<sub>90</sub>',latex_symbol='V_{90}',
                        description='1990 conventional definition of voltage, electrical potential')
    _ohm90 = PrefixedUnit('ohm 90','\u03A9(90)',Conversion(_ohm,RK/RK90),add_symbol=True,
                          html_symbol='&Omega;<sub>90</sub>',latex_symbol=r'\Omega_{90}',ascii_symbol='ohm(90)',
                          description='1990 conventional definition of electrical resistance')
    _A90 = PrefixedUnit('ampere 90','A(90)',Conversion(_V90*_ohm90**-1,1),add_symbol=True,
                        html_symbol='A<sub>90</sub>',latex_symbol='A_{90}',
                        description='1990 conventional definition of electrical current')
    PrefixedUnit('coulomb 90','C(90)',Conversion(_s*_V90*_ohm90**-1,1),add_symbol=True,
                 html_symbol='C<sub>90</sub>',latex_symbol='C_{90}',
                 description='1990 conventional definition of electrical charge')
    PrefixedUnit('watt 90','W(90)',Conversion(_V90*_A90,1),add_symbol=True,
                 html_symbol='W<sub>90</sub>',latex_symbol='W_{90}',
                 description='1990 conventional definition of electrical power')
    PrefixedUnit('farad 90','F(90)',Conversion(_s*_ohm90**-1,1),add_symbol=True,
                 html_symbol='F<sub>90</sub>',latex_symbol='F_{90}',
                 description='1990 conventional definition of electrical capacitance')
    PrefixedUnit('henry 90','H(90)',Conversion(_s*_ohm90,1),add_symbol=True,
                 html_symbol='H<sub>90</sub>',latex_symbol='H_{90}',
                 description='1990 conventional definition of electrical inductance')
    
    
    # 1954 Kelvin definition
    PrefixedUnit('kelvin 54','K(54)',Conversion(_K,gummy(1,u=3.7e-7)),add_symbol=True,
                 html_symbol='K<sub>54</sub>',latex_symbol='K_{54}',
                 description = 'unit of temperature, 1954 SI definition based on the triple point of water')




# -*- coding: utf-8 -*-

# module test_misc

# Copyright (C) 2025 National Research Council Canada
# Author:  Harold Parks

# This file is part of MetroloPy.

# MetroloPy is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software 
# Foundation, either version 3 of the License, or (at your option) any later 
# version.

# MetroloPy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more 
# details.

# You should have received a copy of the GNU General Public License along with 
# MetroloPy. If not, see <http://www.gnu.org/licenses/>.
"""
Simple unit tests for the gummy module
"""

import metrolopy as uc
import numpy as np
import unittest

class TestGummy(unittest.TestCase):
    """
    Various unit tests of gummy properties and methods.
    """
    def test_ku(self):
        from metrolopy.gummy import _ku
        from decimal import Decimal
        
        self.assertTrue(_ku(2.1,3.3) == 2.1*3.3)
        self.assertTrue(_ku(2.1,Decimal(3.3)) == Decimal(float(2.1))*Decimal(3.3))
        
    def test_meta_properties(self):
        """
        Test class properties.

        """
        a = uc.gummy.cimethod
        self.assertTrue(isinstance(a,str))
        uc.gummy.cimethod = 'symmetric'
        self.assertTrue(uc.gummy.cimethod == 'symmetric')
        g = uc.gummy(1,1)
        self.assertTrue(g.cimethod == 'symmetric')
        
        try:
            uc.gummy.cimethod = 'asdfkjhasd'
            self.assertTrue(False)
        except ValueError:
            pass
        
        uc.gummy.cimethod = a
    
        b = uc.gummy.bayesian
        self.assertTrue(isinstance(b,bool))
        uc.gummy.bayesian = not b
        self.assertTrue(uc.gummy.bayesian is not b)
        g = uc.gummy(1,1)
        self.assertTrue(g.bayesian is not b)
        
        uc.gummy.bayesian = b
        
        b = uc.gummy.style
        self.assertTrue(isinstance(b,str))
        uc.gummy.style = 'xf'
        self.assertTrue(uc.gummy.style == 'xf')
        g = uc.gummy(1,1)
        self.assertTrue(g.style == 'xf')
        
        try:
            uc.gummy.style = 'asdfjkhasdf'
            self.assertTrue(False)
        except ValueError:
            pass
        
        uc.gummy.style = b
        
        k = uc.gummy.cmp_k
        p = uc.gummy.cmp_p
        if p is None:
            self.assertTrue(k >= 0)
        else:
            self.assertTrue(k is None)
            self.assertTrue(p >= 0)
            self.assertTrue(p <= 1)
            
        
        uc.gummy.cmp_k = 2.1
        self.assertTrue(uc.gummy.cmp_p is None)
        self.assertTrue(uc.gummy.cmp_k == 2.1)
        
        if k is None:
            uc.gummy.cmp_p = p
        else:
            uc.gummy.cmp_k = k
            
        try:
            uc.gummy.p_method = 'ccp'
            self.assertTrue(uc.gummy.p_method == 'ccp')
            uc.gummy.p_method = None
            self.assertTrue(uc.gummy.p_method == 'loc')
        finally:
            uc.gummy.p_method = 'loc'
        
        d = uc.gummy.max_dof
        uc.gummy.max_dof = 1021
        self.assertTrue(uc.gummy.max_dof == 1021)
        uc.gummy.max_dof = d
        
        n = uc.gummy.nsig
        uc.gummy.nsig = n + 1
        self.assertTrue(uc.gummy.nsig == n + 1)
        g = uc.gummy(1,1)
        self.assertTrue(g.nsig == n + 1)
        uc.gummy.nsig = n
        
        ts = uc.gummy.thousand_spaces
        self.assertTrue(isinstance(ts,bool))
        uc.gummy.thousand_spaces = not ts
        self.assertTrue(uc.gummy.thousand_spaces is not ts)
        g = uc.gummy(1,1)
        self.assertTrue(g.thousand_spaces is not ts)
        uc.gummy.thousand_spaces = ts
        
    def test_init(self):
        """
        Test gummy creation
        """
        g = uc.gummy(1.2,3.4)
        self.assertTrue(g.x == 1.2)
        self.assertTrue(g.u == 3.4)
        self.assertTrue(g.U == g.u)
        self.assertTrue(g.unit is uc.one)
        
        gg = uc.gummy(g)
        self.assertTrue(gg is not g)
        self.assertTrue(gg.x == 1.2)
        self.assertTrue(gg.u == 3.4)
        self.assertTrue(g.correlation(gg) == 1)
        
        q = uc.Quantity(5.6,unit='m')
        gq = uc.gummy(q)
        self.assertTrue(gq.x == 5.6)
        self.assertTrue(gq.u == 0)
        self.assertTrue(gq.unit is uc.unit('m'))
        
        u = uc.ummy(7.8,9.1)
        gu = uc.gummy(u)
        self.assertTrue(gu.x == 7.8)
        self.assertTrue(gu.u == 9.1)
        self.assertTrue(gu.unit is uc.one)
        
        ung = uc.gummy(2.1,unit='mm')
        gun = uc.gummy(3.4,u=ung,unit='m')
        self.assertTrue(gun.unit is uc.unit('m'))
        self.assertTrue(gun.x == 3.4)
        self.assertTrue(gun.U == 2.1)
        self.assertTrue(gun.u == 2.1/1000)
        self.assertTrue(gun.uunit is uc.unit('mm'))
        
        ung = uc.Quantity(2.1,unit='mm')
        gun = uc.gummy(3.4,u=ung,unit='m')
        self.assertTrue(gun.unit is uc.unit('m'))
        self.assertTrue(gun.x == 3.4)
        self.assertTrue(gun.U == 2.1)
        self.assertTrue(gun.u == 2.1/1000)
        self.assertTrue(gun.uunit is uc.unit('mm'))
        
        guu = uc.gummy(1,1,unit='m',uunit='m')
        self.assertTrue(guu.unit is uc.unit('m'))
        self.assertTrue(guu.uunit is None)
        
        guuz = uc.gummy(1,unit='m',uunit='mm')
        self.assertTrue(guuz.unit is uc.unit('m'))
        self.assertTrue(guuz.uunit is None)
        
        gp = uc.gummy(2,1,p=0.95)
        self.assertTrue(gp.x == 2)
        self.assertTrue(gp.p == 0.95)
        self.assertTrue(abs(gp.k - 1.96) < 0.01)
        self.assertTrue(abs(gp.u - 1/1.96) < 0.01)
    
        gk = uc.gummy(3,1,k=2)
        self.assertTrue(gk.x == 3)
        self.assertTrue(gk.k == 2)
        self.assertTrue(abs(gk.p - 0.954) < 0.1)
        
        gk = uc.gummy(3,1,k=1.96)
        self.assertTrue(gk.x == 3)
        self.assertTrue(gk.k == 1.96)
        self.assertTrue(abs(gk.p - 0.95) < 0.1)
        
        gd = uc.gummy(uc.TriangularDist(2.2,0.73))
        self.assertTrue(gd.x == 2.2)
        self.assertTrue(abs(gd.u - np.sqrt(0.73**2/6)) < 1e-12)
        self.assertTrue(abs(gd.U - np.sqrt(0.73**2/6)) < 1e-12)
        
        gd = uc.gummy(uc.TriangularDist(2.2,0.73),k=2.6)
        self.assertTrue(gd.x == 2.2)
        self.assertTrue(abs(gd.u - np.sqrt(0.73**2/6)) < 1e-12)
        self.assertTrue(abs(gd.U - 2.6*np.sqrt(0.73**2/6)) < 1e-12)
        
        gd = uc.gummy(uc.TriangularDist(2.2,0.73),k=2.6,unit='m',uunit='cm')
        self.assertTrue(gd.x == 2.2)
        self.assertTrue(abs(gd.u - np.sqrt(0.73**2/6)) < 1e-12)
        self.assertTrue(abs(gd.U - 2.6*100*np.sqrt(0.73**2/6)) < 1e-12)
        
        try:
            uc.gummy(3.3,1.3,unit='%',uunit='m')
            self.assertTrue(False)
        except uc.IncompatibleUnitsError:
            pass
        
        gg = uc.gummy(3.3,1.4,unit='%',uunit=uc.one)
        self.assertTrue(abs(gg.u - 100*1.4) < 1e-12)
        
        gg = uc.gummy(3.3,1.4,unit='%',uunit='ppm')
        self.assertTrue(abs(gg.u - 3.3*1.4e-6) < 1e-12)
        
        gg = uc.gummy(3.3,1.4,unit='m',uunit='%')
        self.assertTrue(abs(gg.u - 3.3*0.014) < 1e-12)
        
        gg = uc.gummy(1,1,dof=3.3,utype='abc',name='def')
        self.assertTrue(gg.dof == 3.3)
        self.assertTrue(gg.utype == 'abc')
        self.assertTrue(gg.name == 'def')
        
    def test_U(self):
        """
        Test expanded uncertainty.
        """
        a = uc.gummy(1.1,2.2)
        self.assertTrue(a.U == 2.2)
        b = uc.gummy(3,4.4)
        c = a + b
        c.ubreakdown = [a,b]
        self.assertTrue(c.U[0] == 2.2)
        self.assertTrue(c.U[1] == 4.4)
        
        a = uc.gummy(1.1,2.2,'m')
        self.assertTrue(a.U == 2.2)
        b = uc.gummy(3,4.4,'s')
        c = a/b
        c.ubreakdown = [a,b]
        self.assertTrue(abs(c.U[0] - 2.2/3) < 1e-12)
        self.assertTrue(abs(c.U[1] - 4.4*1.1/3**2) < 1e-12)
        
    def test_set_U(self):
        """
        Additional tests for expanded uncertainty.
        """
        g = uc.gummy(1.0,1.0,k=2,unit='cm',uunit='mm')
        self.assertTrue(g.U == 1)
        self.assertTrue(g.uunit is uc.unit('mm'))
        g.unit = 'm'
        self.assertTrue(g.U == 0.001)
        
        a = uc.gummy(1.1,2.2)
        b = uc.gummy(3,4.4)
        c = a + b
        c.ubreakdown = [a,b]
        self.assertTrue(c.U[0] == 2.2)
        self.assertTrue(c.U[1] == 4.4)
        
    def test_iset_U(self):
        """
        Additional tests for expanded uncertainty.
        """
        g = uc.gummy(1)
        self.assertTrue(g._iset_U() == 0)
        self.assertTrue(g._iset_U(unit='m').unit is uc.unit('m'))
        self.assertTrue(g._iset_U(unit='m').value == 0)
        self.assertTrue(g._iset_U(u=float('inf')) == float('inf'))
        self.assertTrue(g._iset_U(u=float('inf'),unit='m').value == float('inf'))
        self.assertTrue(g._iset_U(u=float('inf'),unit='m').unit is uc.unit('m'))
        
        g = uc.gummy(1,1,k=2,unit='m')
        self.assertTrue(g._iset_U(unit='m').value == 1)
        self.assertTrue(g._iset_U(k=4) == 2)
        self.assertTrue(g._iset_U(unit='%').value == 100)
        self.assertTrue(g._iset_U(unit='%').unit is uc.unit('%'))
        self.assertTrue(g._iset_U(unit='cm').value == 100)
        self.assertTrue(g._iset_U(unit='cm').unit is uc.unit('cm'))
        
        try:
            self.assertTrue(g._iset_U(unit='kg'))
            self.assertTrue(False)
        except uc.IncompatibleUnitsError:
            pass
        
        g = uc.gummy(0,1,unit='m')
        g._iset_U(unit='%').value == float('inf')
        g._iset_U(unit='%').unit is uc.unit('%')
        
        g = uc.gummy(1,1,unit='degC')
        self.assertTrue(g._iset_U(k=2) == 2)
        
    def test_Usim(self):
        """
        Test the gummy.Usim property
        """
        g = uc.gummy(-8.5,1,unit='m')
        g.sim()
        self.assertTrue(abs(g.Usim[0] - 1) < 0.5)
        self.assertTrue(abs(g.Usim[1] - 1) < 0.5)
        g.uunit = '%'
        self.assertTrue(abs(g.Usim[0] - 100*1/8.5) < 10)
        self.assertTrue(abs(g.Usim[1] - 100*1/8.5) < 10)
        g.uunit = 'mm'
        self.assertTrue(abs(g.Usim[0] - 1000) < 500)
        self.assertTrue(abs(g.Usim[1] - 1000) < 500)
        
    def test_xusim(self):
        """
        tests gummy.xsim and gummy.usim
        """
        g = uc.gummy(-3.3,0.11)
        g.sim()
        self.assertTrue(abs(g.xsim + 3.3) < 0.1)
        self.assertTrue(abs(g.usim - 0.11) < 0.1)
        
    def test_cisim(self):
        """
        tests gummy.cisim and gummy.cimethod
        """
        g = uc.gummy(-3.3,0.11)
        g.sim()
        
        g.cimethod = 'symmetric'
        ci = g.cisim
        self.assertTrue(abs(ci[0] + 3.41) < 0.1)
        self.assertTrue(abs(ci[1] + 3.19) < 0.1)
        
        g.cimethod = 'shortest'
        ci = g.cisim
        self.assertTrue(abs(ci[0] + 3.41) < 0.1)
        self.assertTrue(abs(ci[1] + 3.19) < 0.1)
        
        try:
            g.cimethod = 'asdfhasd'
            self.assertTrue(False)
        except ValueError:
            pass
        
        a = uc.gummy(uc.UniformDist(center=0,half_width=1))
        b = a**2
        b.p = 0.8
        b.cimethod = 'shortest'
        b.sim()
        ci = b.cisim
        self.assertTrue(abs(ci[0]) < 0.001)
        self.assertTrue(abs(ci[1] - 0.64) < 0.01)
        b.cimethod = 'symmetric'
        ci = b.cisim
        self.assertTrue(abs(ci[0]- 0.01) < 0.001)
        self.assertTrue(abs(ci[1] - 0.81) < 0.01)
        
    def test_simdata(self):
        """
        tests gummy.simdata and gummy.simsorted
        """
        g = uc.gummy(1,1)
        g.sim(n=10)
        self.assertTrue(len(g.simdata) == 10)
        self.assertTrue(len(g.simsorted) == 10)
        for i in range(9):
            self.assertTrue(g.simsorted[i+1] >= g.simsorted[i])
            
    def test_distribution(self):
        """
        Test gummy.distribution
        """
        g = uc.gummy(1,1)
        self.assertTrue(isinstance(g.distribution,uc.Distribution))
       
    def test_ksim(self):
        """
        Test gummy.ksim
        """
        g = uc.gummy(1,1)
        g.p = 0.9545
        g.sim()
        self.assertTrue(abs(g.ksim - 2) < 0.1)
        
    def test_independent(self):
        """
        Test gummy.independant
        """
        a = uc.gummy(1,1)
        b = uc.gummy(-2,2)
        c = a + b
        self.assertTrue(a.independent)
        self.assertTrue(not c.independent)
        d = uc.gummy(1)
        self.assertTrue(not d.independent)
        
    def test_name(self):
        """
        Test setting and getting gummy.name
        """
        g = uc.gummy(1,1)
        self.assertTrue(g.name is None)
        g.name = 'abc'
        self.assertTrue(g.name == 'abc')
        g.name = ['def','g','hi','jkl']
        self.assertTrue(g.name == 'def')
        self.assertTrue(g.get_name() == 'def')
        self.assertTrue(g.get_name(fmt='html') == 'g')
        self.assertTrue(g.get_name(fmt='latex') == 'hi')
        self.assertTrue(g.get_name(fmt='ascii') == 'jkl')
        g.name = 'f'
        self.assertTrue(g.get_name(fmt='html') == '<i>f</i>')
        g.name = 'ff'
        self.assertTrue(g.get_name(fmt='latex') == '\\text{ff}')
        self.assertTrue(g.get_name(fmt='latex',norm=lambda x:'*'+x+'*') == '*ff*')
        
        try:
            g.get_name(fmt='asdgfads')
            self.assertTrue(False)
        except ValueError:
            pass
        
        g.name = None
        self.assertTrue(g.name is None)
        
        try:
            g.name = 3
            self.assertTrue(False)
        except ValueError:
            pass
        
        try:
            g.name = ['a','b']
            self.assertTrue(False)
        except ValueError:
            pass
        
        try:
            g.name = [1,2]
            self.assertTrue(False)
        except ValueError:
            pass
        
    def test_unit(self):
        """
        Test setting and getting gummy.unit
        """
        g = uc.gummy(1,1,unit='m')
        self.assertTrue(g.unit is uc.unit('m'))
        self.assertTrue(str(g.unit) == 'm')
        g.unit = 'cm'
        self.assertTrue(str(g.unit) == 'cm')
        self.assertTrue(g.x == 100)
        self.assertTrue(g.u == 100)
        g.unit = uc.unit('mm')
        self.assertTrue(str(g.unit) == 'mm')
        self.assertTrue(g.x == 1000)
        self.assertTrue(g.u == 1000)
        
    def test_uunit(self):
        """
        Test setting and getting gummy.uunit
        """
        g = uc.gummy(1,1)
        self.assertTrue(g.uunit is None)
        g = uc.gummy(1,1,unit='m',uunit='cm')
        self.assertTrue(str(g.uunit) == 'cm')
        
        g = uc.gummy(1,unit='m',uunit='cm')
        g.uunit = 'mm'
        self.assertTrue(g.uunit is None)
        
        g = uc.gummy(1,1,unit='m')
        self.assertTrue(g.uunit is None)
        g.uunit = 'cm'
        self.assertTrue(str(g.uunit) == 'cm')
        self.assertTrue(g.U == 100)
        g.uunit = None
        self.assertTrue(g.uunit is None)
        
    def test_uunit_is_rel(self):
        """
        Test gummy.uunit_is_rel
        """
        g = uc.gummy(1,1)
        self.assertTrue(not g.uunit_is_rel)
        g.uunit='%'
        self.assertTrue(g.uunit_is_rel)
        
        g = uc.gummy(1,1,unit='m')
        self.assertTrue(not g.uunit_is_rel)
        g.uunit = 'mm'
        self.assertTrue(not g.uunit_is_rel)
        g.uunit = '%'
        self.assertTrue(g.uunit_is_rel)
        
    def test_k(self):
        """
        Test setting and getting gummy.k
        """
        g = uc.gummy(1)
        self.assertTrue(g.k is None)
        
        g = uc.gummy(1,1,k=2)
        self.assertTrue(g.k == 2)
        g.k = 3
        self.assertTrue(g.k == 3)
        try:
            g.k = -1
            self.assertTrue(False)
        except ValueError:
            pass
        
        uc.gummy.p_method = None
        g.p = 0.68
        self.assertTrue(abs(g.k - 0.99) < 0.01)
        g.p = 0.95
        self.assertTrue(abs(g.k - 1.96) < 0.01)
        
    def test_p(self):
        """
        Test setting and getting gummy.p
        """
        uc.gummy.p_method = None
        g = uc.gummy(1,1)
        g.p = 0.95
        self.assertTrue(g.p == 0.95)
        g.k = 2
        self.assertTrue(abs(g.p - 0.9545) < 0.0001)
        
        try:
            g.p = -0.1
            self.assertTrue(False)
        except ValueError:
            pass
        try:
            g.p = 1.1
            self.assertTrue(False)
        except ValueError:
            pass
        
    def test_correlation(self):
        """
        Test gummy.correlation
        """
        a = uc.gummy(1,1)
        b = uc.gummy(1,2)
        c = a + b
        self.assertTrue(a.correlation(b) == 0)
        self.assertTrue(a.correlation(1) == 0)
        self.assertTrue(abs(c.correlation(a) - 1/c.u) < 1e-4)
        self.assertTrue(abs(a.correlation(c) - 1/c.u) < 1e-4)
        self.assertTrue(abs(c.correlation(b) - 2/c.u) < 1e-4)
        self.assertTrue(abs(b.correlation(c) - 2/c.u) < 1e-4)
        self.assertTrue(a.correlation(a) == 1)
        
    def test_covariance(self):
        """
        Test gummy.covariance
        """
        a = uc.gummy(1,1)
        b = uc.gummy(1,2)
        c = a + b
        self.assertTrue(a.covariance(b) == 0)
        self.assertTrue(a.covariance(1) == 0)
        self.assertTrue(abs(c.covariance(a) - 1) < 1e-4)
        self.assertTrue(abs(a.covariance(c) - 1) < 1e-4)
        self.assertTrue(abs(c.covariance(b) - 4) < 1e-4)
        self.assertTrue(abs(b.covariance(c) - 4) < 1e-4)
        self.assertTrue(a.covariance(a) == 1)
        self.assertTrue(b.covariance(b) == 4)
        
    def test_correlation_sim(self):
        """
        Test gummy.correlation_sim
        """
        a = uc.gummy(1,1)
        b = uc.gummy(1,2)
        c = a + b
        uc.gummy.simulate([a,b,c])
        self.assertTrue(abs(a.correlation_sim(b)) < 1e-1)
        self.assertTrue(abs(c.correlation_sim(a) - 1/c.u) < 1e-1)
        self.assertTrue(abs(a.correlation_sim(c) - 1/c.u) < 1e-1)
        self.assertTrue(abs(c.correlation_sim(b) - 2/c.u) < 1e-1)
        self.assertTrue(abs(b.correlation_sim(c) - 2/c.u) < 1e-1)
        self.assertTrue(abs(a.correlation_sim(a) - 1) < 1e-6)
        
    def test_covariance_sim(self):
        """
        Test gummy.covariance_sim
        """
        a = uc.gummy(1,1)
        b = uc.gummy(1,2)
        c = a + b
        uc.gummy.simulate([a,b,c])
        self.assertTrue(abs(a.covariance_sim(b)) < 1e-1)
        self.assertTrue(abs(c.covariance_sim(a) - 1) < 1e-1)
        self.assertTrue(abs(a.covariance_sim(c) - 1) < 1e-1)
        self.assertTrue(abs(c.covariance_sim(b) - 4) < 1e-1)
        self.assertTrue(abs(b.covariance_sim(c) - 4) < 1e-1)
        self.assertTrue(abs(a.covariance_sim(a) - 1) < 1e-1)
        self.assertTrue(abs(b.covariance_sim(b) - 4) < 1e-1)
        
    def test_correlation_matrix(self):
        """
        Test gummy.correlation_matrix
        """
        a = uc.gummy(1,1)
        b = uc.gummy(1,2)
        c = a + b
        uc.gummy.simulate([a,b,c])
        m = uc.gummy.correlation_matrix_sim([a,b,c])
        self.assertTrue(abs(m[0][0] - 1) < 1e-1)
        self.assertTrue(abs(m[1][1] - 1) < 1e-1)
        self.assertTrue(abs(m[2][2] - 1) < 1e-1)
        self.assertTrue(abs(m[0][1]) < 1e-1)
        self.assertTrue(abs(m[1][0]) < 1e-1)
        self.assertTrue((m[2][0] - 1/c.u) < 1e-1)
        self.assertTrue((m[0][2] - 1/c.u) < 1e-1)
        self.assertTrue((m[2][1] - 2/c.u) < 1e-1)
        self.assertTrue((m[1][2] - 2/c.u) < 1e-1)
        
    def test_covariance_matrix(self):
        """
        Test gummy.covariance_matrix
        """
        a = uc.gummy(1,1)
        b = uc.gummy(1,2)
        c = a + b
        uc.gummy.simulate([a,b,c])
        m = uc.gummy.covariance_matrix_sim([a,b,c])
        self.assertTrue(abs(m[0][0] - 1) < 1e-1)
        self.assertTrue(abs(m[1][1] - 4) < 1e-1)
        self.assertTrue(abs(m[2][2] - 5) < 1e-1)
        self.assertTrue(abs(m[0][1]) < 1e-1)
        self.assertTrue(abs(m[1][0]) < 1e-1)
        self.assertTrue((m[2][0] - 1) < 1e-1)
        self.assertTrue((m[0][2] - 1) < 1e-1)
        self.assertTrue((m[2][1] - 4) < 1e-1)
        self.assertTrue((m[1][2] - 4) < 1e-1)
        
    def test_finfo(self):
        """
        Test gummy.finfo
        """
        a = uc.gummy(1,1)
        self.assertTrue(a.finfo.rel_u == 0)
        
    def test_real(self):
        """
        Test gummy.real
        """
        a = uc.gummy(1.1,0.3)
        b = a.real
        self.assertTrue(not a is b)
        self.assertTrue(a.x == b.x)
        self.assertTrue(a.u == b.u)
        self.assertTrue(a.correlation(b) == 1)
        
    def test_conjugate(self):
        """
        Test gummy.conjugate
        """
        a = uc.gummy(1.1,0.3)
        b = a.conjugate()
        self.assertTrue(not a is b)
        self.assertTrue(a.x == b.x)
        self.assertTrue(a.u == b.u)
        self.assertTrue(a.correlation(b) == 1)
        
    def test_angle(self):
        """
        Test gummy.angle
        """
        self.assertTrue(uc.gummy(1.1,0.3).angle() == 0)
        self.assertTrue(abs(abs(uc.gummy(-1.1,0.3).angle()) - np.pi) < 1e-10)
        
    def test_utype(self):
        """
        Test gummy.utype
        """
        a = uc.gummy(1.2,1,1,utype='A')
        self.assertTrue(a.utype == 'A')
        b = uc.gummy(1.2,1,1,utype='xyz')
        self.assertTrue(b.utype == 'xyz')
        
    def test_ufrom(self):
        """
        Test gummy.ufrom
        """
        a = uc.gummy(1.2,0.2,utype='A')
        b = uc.gummy(3.2,0.5,utype='A')
        c = uc.gummy(0.9,0.2,utype='B')
        d = a + b + c
        self.assertTrue(abs(d.ufrom('A') - np.sqrt(0.2**2+0.5**2)) < 1e-8)
        uc.gummy.simulate([d,b])
        self.assertTrue(abs(d.ufromsim('A') - np.sqrt(0.2**2+0.5**2)) < 1e-2)
        self.assertTrue(d.ufrom(c) == 0.2)
        
    def test_dof_from(self):
        """
        Test gummy.doffrom
        """
        a = uc.gummy(1.2,0.2,dof=5,utype='A')
        b = uc.gummy(3.2,0.5,dof=7,utype='A')
        c = uc.gummy(0.9,0.2,utype='B')
        d = a + b + c
        self.assertTrue(abs(d.doffrom('A') - (0.2**2+0.5**2)**2/(0.2**4/5 + 0.5**4/7)) < 1e-10)
        self.assertTrue(d.doffrom(a) == 5)
        
            
if __name__ == '__main__':
    unittest.main()


# -*- coding: utf-8 -*-

# module unitparser

# Copyright (C) 2025 National Research Council Canada
# Author:  Harold Parks

# This file is part of MetroloPy.

# MetroloPy is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software 
# Foundation, either version 3 of the License, or (at your option) any later 
# version.

# MetroloPy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more 
# details.

# You should have received a copy of the GNU General Public License along with 
# MetroloPy. If not, see <http://www.gnu.org/licenses/>.

"""
unit parser
"""

from .exceptions import UnitLibError
import numpy as np
from fractions import Fraction

def _get_num(txt):
    s = 1
    txt = txt.strip()
    if txt.startswith('-'):
        s = -s
        txt = txt[1:]
    elif txt.startswith('+'):
        txt = txt[1:]
    txt = txt.lstrip(' (').rstrip(' )')
    if txt.startswith('-'):
        s = -s
        txt = txt[1:]
    elif txt.startswith('+'):
        txt = txt[1:]
    f = float(txt)
    if np.modf(f)[0] == 0:
        return s*int(f)
    else:
        return s*f
    
def _get_exp(txt):
    txt = txt.split('/')
    if len(txt) == 1:
        return _get_num(txt[0])
    elif len(txt) == 2:
        return Fraction(_get_num(txt[0]),_get_num(txt[1]))
    else:
        raise UnitLibError('syntax error in exponent: ' + txt)

class _UnitParser:
    def __init__(self,txt):  
        self.txt= txt
        self.n = len(txt)
        self.i = 0
        
    def parse(self):
        return self.ppar(0,1)
                
    def ppar(self,lev,es):
        cu = ''
        ul = []
        sol = False
        while self.i < self.n:
            c = self.txt[self.i]
            if c == ')':
                cu = cu.strip()
                if cu != '':
                    ul += [[cu,es]]
                if len(ul) == 0:
                    raise UnitLibError('syntax error in: ' + self.txt)
                self.i += 1
                e = self.look_for_pow()
                if e != 1:
                    for u in ul:
                        u[1] *= e
                        if np.modf(u[1]) == 0:
                            u[1] = int(u[1])
                return ul
            elif c == '(':
                cu = cu.strip()
                p = self.pch(cu,')',']',self.i)
                if p is None:
                    if cu != '':
                        ul += [[cu,es]]
                        cu = ''
                    self.i += 1
                    ul += self.ppar(lev+1,es)
                    if sol:
                        es = -es
                        sol = False
                else:
                    cu += p
                    self.i += len(p)
            elif c == '[':
                cu = cu.strip()
                p = self.pch(cu,']',')',self.i)
                if p is None:
                    if cu != '':
                        ul += [[cu,es]]
                        cu = ''
                    self.i += 1
                    ul += [self.pbr(es,cu)]
                    cu = ''
                    if sol:
                        es = -es
                        sol = False
                else:
                    cu += p
                    self.i += len(p)
            elif c.isspace() or c == '*':
                cu = cu.strip()
                if cu != '':
                    ul += [[cu,es*self.look_for_pow()]]
                cu = ''
                if sol:
                    es = -es
                    sol = False
            elif c == '/':
                cu = cu.strip()
                if cu != '':
                    ul += [[cu,es]]
                    cu = ''
                 #if sol:
                    # raise UnitLibError('ambugus use of several soliudses (forward slashes) in: ' + self.txt)
                sol = not sol
                es = -es
                self.i += 1
            else:
                cu += c
                self.i += 1
        if lev != 0:
            raise UnitLibError('unmatched ")" found in: ' + self.txt)
        cu = cu.strip()
        if cu != '':
            ul += [[cu,es]]
        return ul
            
    def look_for_pow(self):
        while self.i < self.n:
            c = self.txt[self.i]
            if c == '*':
                if self.i == (self.n - 1):
                    raise UnitLibError(self.txt + ' ends with *')
                self.i += 1
                c = self.txt[self.i]
                if not c == '*':
                    return 1
                par = -1
                et = ''
                self.i += 1
                cl = ''
                stnum = False
                while self.i < self.n:
                    c = self.txt[self.i]
                    if c == '(':
                        if par == 0:
                            if cl.isspace() or cl == '*':
                                return _get_exp(et)
                            raise UnitLibError('syntax error in: ' + self.txt)
                        else:
                            if par == -1:
                                par = 1
                            else:
                                par += 1
                    elif c == '[':
                        if cl.isspace() or cl == '*':
                            return _get_exp(et)
                        raise UnitLibError('syntax error in: ' + self.txt )
                    elif c == ']':
                        raise UnitLibError('unmatched bracket: ' + self.txt)
                    elif c.isalpha():
                        if par > 0:
                            raise UnitLibError('syntax error in: ' + self.txt)
                        return _get_exp(et)
                    elif c == '/':
                        if par <= 0:
                            return _get_exp(et)
                    elif c == ')':
                        if par <= 0:
                            return _get_exp(et)
                        else:
                            par -= 1
                    elif c.isnumeric() or c =='.':
                        stnum = True
                    elif c == '+'  or c == '-':
                        if stnum and par <= 0:
                            raise UnitLibError('syntax error in: ' + self.txt)
                    elif c == '*':
                        if self.i == self.n - 1:
                            raise UnitLibError('syntax error in: ' + self.txt)
                        if par <= 0 and cl != '*' and self.txt[self.i+1] != '*':
                            self.i += 1
                            return _get_exp(et)
                    elif c.isspace():
                        if stnum and par <= 0:
                            self.i += 1
                            return _get_exp(et)
                    else:
                        if par <= 0:
                            raise UnitLibError('syntax error in: ' + self.txt)
                    self.i += 1
                    cl = c
                    et += c
                return _get_exp(et)
    
            elif not c.isspace():
                return 1
            self.i += 1
        return 1
            
    def pbr(self,es,cu):
        br = 0
        while self.i < self.n:
            c = self.txt[self.i]
            if c == '[':
                br += 1
                cu += c
            if c == ']':
                if br == 0:
                    self.i += 1
                    return [cu,es*self.look_for_pow()]
                br -= 1
            cu += c
            self.i += 1
        raise UnitLibError('syntax error in: ' + self.txt)
        
    def pch(self,cu,ech,och,i):
        if cu == '':
            return None
        ret = self.txt[i]
        i += 1
        while i < self.n:
            c = self.txt[i]
            
            if c == ech:
                return ret + c
            if c.isspace():
                return None
            if c in {och,'*','/'}:
                return None
            
            if c == '(':
                r = self.pch('a',')',']',i)
                if r is None:
                    return None
                else:
                    ret += r
                    i += len(r) - 1
            elif c == '[':
                r = self.pch('a',']',')',i)
                if r is None:
                    return None
                else:
                    ret += r
                    i += len(r) - 1
            else:
                ret += c
                
            i += 1
            
        return None
        
        